@@ -37,6 +37,6 @@ extern int consolePosition;
 extern GQueue* consoleLines;
 
 void addConsoleLine(char* text, int r, int g, int b);
-void deleteOldLines();
+void deleteOldConsoleLines();
 
 #endif /* CONSOLE_H_ */
