@@ -12,11 +12,7 @@
  */
 
 #ifdef N900
-#include <GLES/gl.h>
-#include <GLES/glext.h>
 #include <SDL/SDL_gles.h>
-#else
-#include <GL/gl.h>
 #endif
 
 #include <SDL/SDL.h>
@@ -25,25 +21,39 @@
 #include "tileengine.h"
 #include "glunproject.h"
 #include "texture.h"
+#include "markerIterator.h"
 
-#define MAX_SEARCH_MARKERS 20
-#define VERTICES_PER_SEARCH_MARKER 4
-//#define VERTICES_PER_SEARCH_MARKER_SHADOW 3
-#define VERTEX_INDICES_PER_SEARCH_MARKER 6
+#define MAX_HOLD_INDICATOR_POINTS 12
+#define MAX_MARKERS 100
+#define MAX_ROUTE_SEGMENTS 1000
+#define QUADS_PER_ROUTE_SEGMENT 3
+#define MAX_ROUTE_QUADS (QUADS_PER_ROUTE_SEGMENT*MAX_ROUTE_SEGMENTS)
+#define MAX_QUADS MAX_ROUTE_QUADS
 #define FLOATS_PER_VERTICE 3
-//#define FLOATS_PER_COLOR 4
-//GLfloat searchMarkerShadowVertices[MAX_SEARCH_MARKERS * VERTICES_PER_SEARCH_MARKER_SHADOW * FLOATS_PER_VERTICE];
-//GLfloat searchMarkerShadowColors[MAX_SEARCH_MARKERS * VERTICES_PER_SEARCH_MARKER_SHADOW * FLOATS_PER_COLOR];
-GLfloat searchMarkerVertices[MAX_SEARCH_MARKERS * VERTICES_PER_SEARCH_MARKER * FLOATS_PER_VERTICE];
-GLushort searchMarkerVertexIndices[MAX_SEARCH_MARKERS * VERTEX_INDICES_PER_SEARCH_MARKER];
-GLfloat searchMarkerTextureCoords[MAX_SEARCH_MARKERS * VERTICES_PER_SEARCH_MARKER * 2];
+#define VERTICES_PER_QUAD 4
+#define FLOATS_PER_QUAD (FLOATS_PER_VERTICE*VERTICES_PER_QUAD)
+#define VERTEX_INDICES_PER_QUAD 6
+GLfloat mouseHoldIndicatorVertices[MAX_HOLD_INDICATOR_POINTS * FLOATS_PER_QUAD];
+
+GLfloat searchMarkerVertices[MAX_MARKERS * FLOATS_PER_QUAD];
+GLfloat shadowVertices[MAX_MARKERS * VERTICES_PER_QUAD * FLOATS_PER_VERTICE];
+GLushort quadVertexIndices[MAX_QUADS * VERTEX_INDICES_PER_QUAD];
+GLfloat quadTextureCoords[MAX_QUADS * VERTICES_PER_QUAD * 2];
+ScreenMarker* renderedMarkers[MAX_MARKERS];
+
+GLfloat routeVertices[MAX_ROUTE_QUADS * VERTICES_PER_QUAD * FLOATS_PER_VERTICE];
+GLfloat routeTextCoords[MAX_ROUTE_QUADS * VERTICES_PER_QUAD * 2];
 
 #define MAX_BUILDINGS_VERTICES (2000*18)
 GLfloat buildingsVertices[MAX_BUILDINGS_VERTICES];
 int wallCount;
 
-Texture fontMediumMask, fontMedium, barbg, barbgmask;
+Texture fontMediumMask, fontMedium;
 Texture searchMarkTexture, searchMarkMask;
+Texture routeStartTexture, routeStartMask;
+Texture routeEndTexture, routeEndMask;
+Texture shadow, dragCrosshair;
+Texture road;
 
 const int MEDIUM_FONT_METRICS[256] = { 8, 8, 8, 8, 8, 8, 8, 8, 8, 56, 0, 8, 8, 0, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 4, 5, 5, 11, 8, 12, 11, 3, 5, 5, 7, 11, 4, 5, 4, 4, 8, 8, 8, 8,
 		8, 8, 8, 8, 8, 8, 4, 4, 11, 11, 11, 7, 13, 9, 9, 9, 10, 8, 7, 10, 10, 3, 3, 8, 7, 11, 10, 10, 8, 10, 8, 9, 7, 10, 9, 11, 10, 7, 10, 5, 4, 5, 11, 7, 7, 8, 8, 7, 8, 8, 4, 8, 8, 3, 3, 7, 3, 13,
@@ -94,6 +104,11 @@ SDL_Surface* loadTextureFromFile(char *filename, t_tile* tile, SDL_Surface** mas
 GLuint createTexture(GLushort * pixels4444, int w, int h, int isTile);
 
 char strbuf[MAX_BUF_SIZE];
+char strbuf2[MAX_BUF_SIZE];
+
+void resetQuadStripVertices() {
+	memset(quadStripVertices, 0, 12 * sizeof(GLfloat));
+}
 
 void setQuadTriangleSize(GLfloat w, GLfloat h, GLfloat quad[12]) {
 	quad[3] = w;
@@ -201,8 +216,8 @@ void drawStringAlpha(char* str, GLfloat r, GLfloat g, GLfloat b, GLfloat a, int
 		//            fprintf(stderr, "textureCoords[%d] = %g\n", i * 12 + j, stringTexCoords[i * 12 + j]);
 		//        }
 	}
+	glEnable(GL_BLEND);
 	if (useMask) {
-		glEnable(GL_BLEND);
 		glBlendFunc(GL_DST_COLOR, GL_ZERO);
 		int glerror;
 		glerror = glGetError();
@@ -264,22 +279,67 @@ void drawBoundingBox() {
 
 int markersInitialized = FALSE;
 
-void initMarkerIndices() {
+void initQuadIndices(GLushort* indices, int count) {
 	int i;
-	for (i = 0; i < MAX_SEARCH_MARKERS; i++) {
-		searchMarkerVertexIndices[VERTEX_INDICES_PER_SEARCH_MARKER * i + 0] = i * 4 + 0;
-		searchMarkerVertexIndices[VERTEX_INDICES_PER_SEARCH_MARKER * i + 1] = i * 4 + 1;
-		searchMarkerVertexIndices[VERTEX_INDICES_PER_SEARCH_MARKER * i + 2] = i * 4 + 2;
-		searchMarkerVertexIndices[VERTEX_INDICES_PER_SEARCH_MARKER * i + 3] = i * 4 + 1;
-		searchMarkerVertexIndices[VERTEX_INDICES_PER_SEARCH_MARKER * i + 4] = i * 4 + 2;
-		searchMarkerVertexIndices[VERTEX_INDICES_PER_SEARCH_MARKER * i + 5] = i * 4 + 3;
+	for (i = 0; i < count; i++) {
+		indices[VERTEX_INDICES_PER_QUAD * i + 0] = i * 4 + 0;
+		indices[VERTEX_INDICES_PER_QUAD * i + 1] = i * 4 + 1;
+		indices[VERTEX_INDICES_PER_QUAD * i + 2] = i * 4 + 2;
+		indices[VERTEX_INDICES_PER_QUAD * i + 3] = i * 4 + 1;
+		indices[VERTEX_INDICES_PER_QUAD * i + 4] = i * 4 + 2;
+		indices[VERTEX_INDICES_PER_QUAD * i + 5] = i * 4 + 3;
 	}
 }
 
-void initMarkerTextureCoords() {
+void initRoadTextureCoords() {
+	int i, idx = 0;
+	for (i = 0; i < MAX_ROUTE_SEGMENTS; i++) {
+		// left quad
+		routeTextCoords[idx++] = 0.0;
+		routeTextCoords[idx++] = 0.0;
+
+		routeTextCoords[idx++] = 0.0;
+		routeTextCoords[idx++] = 1.0;
+
+		routeTextCoords[idx++] = 0.5;
+		routeTextCoords[idx++] = 0.0;
+
+		routeTextCoords[idx++] = 0.5;
+		routeTextCoords[idx++] = 1.0;
+
+		// center quad
+		routeTextCoords[idx++] = 0.5;
+		routeTextCoords[idx++] = 0.0;
+
+		routeTextCoords[idx++] = 0.5;
+		routeTextCoords[idx++] = 1.0;
+
+		routeTextCoords[idx++] = 0.5;
+		routeTextCoords[idx++] = 0.0;
+
+		routeTextCoords[idx++] = 0.5;
+		routeTextCoords[idx++] = 1.0;
+
+		// right quad
+		routeTextCoords[idx++] = 0.5;
+		routeTextCoords[idx++] = 0.0;
+
+		routeTextCoords[idx++] = 0.5;
+		routeTextCoords[idx++] = 1.0;
+
+		routeTextCoords[idx++] = 1.0;
+		routeTextCoords[idx++] = 0.0;
+
+		routeTextCoords[idx++] = 1.0;
+		routeTextCoords[idx++] = 1.0;
+
+	}
+}
+
+void initQuadTextureCoords() {
 	int i;
-	for (i = 0; i < MAX_SEARCH_MARKERS; i++) {
-		memcpy(&searchMarkerTextureCoords[i * 8], texCoordsLandscape, 8 * sizeof(GLfloat));
+	for (i = 0; i < MAX_QUADS; i++) {
+		memcpy(&quadTextureCoords[i * 8], texCoordsLandscape, 8 * sizeof(GLfloat));
 	}
 }
 
@@ -290,24 +350,6 @@ void inline setColor(GLfloat* array, int idx, GLfloat r, GLfloat g, GLfloat b, G
 	array[idx + 3] = a;
 }
 
-//void initMarkerShadowColors() {
-//	int i;
-//	for (i = 0; i < MAX_SEARCH_MARKERS; i++) {
-//		setColor(searchMarkerShadowColors, i * VERTICES_PER_SEARCH_MARKER_SHADOW * FLOATS_PER_COLOR + 0, 0, 0.0, 0.0, .6);
-//		setColor(searchMarkerShadowColors, i * VERTICES_PER_SEARCH_MARKER_SHADOW * FLOATS_PER_COLOR + FLOATS_PER_COLOR, 0, 0.0, 0.0, 0.01);
-//		setColor(searchMarkerShadowColors, i * VERTICES_PER_SEARCH_MARKER_SHADOW * FLOATS_PER_COLOR + FLOATS_PER_COLOR * 2, 0, 0.0, 0.0, 0.01);
-//	}
-//}
-
-void initMarkers() {
-	if (markersInitialized == FALSE) {
-		initMarkerTextureCoords();
-		initMarkerIndices();
-//		initMarkerShadowColors();
-		markersInitialized = TRUE;
-	}
-}
-
 double inline deg2rad(double degrees) {
 	return degrees * M_PI / 180.0;
 }
@@ -319,84 +361,433 @@ void rotate2d(GLfloat* x, GLfloat* y, double rotation) {
 	*y = ty;
 }
 
-int prepareMarkers() {
-	GList* elem = canvas.markers;
-	int verticeIdx = 0, count = 0, i;
-
-	initMarkers();
-
-	setQuadStripSize(searchMarkTexture.size / canvas.scale, searchMarkTexture.size / canvas.scale, quadStripVertices);
-	GLfloat markerRotation = getMarkerRotation();
+void rotateMarkerStripVertices(double angle, GLfloat rotx, GLfloat roty) {
+	int i;
 
 	for (i = 0; i < 4; i++) {
 		GLfloat x2, y2;
-		x2 = quadStripVertices[i * 3] - searchMarkTexture.size / 2.0 / canvas.scale;
-		y2 = quadStripVertices[i * 3 + 1] - searchMarkTexture.size / canvas.scale;
-		rotate2d(&x2, &y2, markerRotation);
-		quadStripVertices[i * 3] = x2 + searchMarkTexture.size / 2.0 / canvas.scale;
-		quadStripVertices[i * 3 + 1] = y2 + searchMarkTexture.size / canvas.scale;
+		x2 = quadStripVertices[i * 3] - rotx;
+		y2 = quadStripVertices[i * 3 + 1] - roty;
+		rotate2d(&x2, &y2, angle);
+		quadStripVertices[i * 3] = x2 + rotx;
+		quadStripVertices[i * 3 + 1] = y2 + roty;
 	}
+}
 
-	GLfloat normRaiseX = 0.0, normRaiseY = -1.0;
-	rotate2d(&normRaiseX, &normRaiseY, markerRotation);
+int searchMarkersCount, routeMarkersCount;
+int prepareMarkers() {
+	int verticeIdx = 0, count = 0, i;
 
+	setQuadStripSize(searchMarkTexture.size / canvas.scale, searchMarkTexture.size / canvas.scale, quadStripVertices);
+	GLfloat markerRotation = getMarkerRotation();
+	GLfloat size = searchMarkTexture.size / canvas.scale;
+	GLfloat halfSize = size / 2.0;
 
-	while (elem != NULL) {
-		ScreenMarker* marker = (ScreenMarker*) elem -> data;
+	rotateMarkerStripVertices(markerRotation, halfSize, size);
+	Point markerRaise;
+	markerRaise.x = 0.0;
+	markerRaise.y = -1.0;
+	rotate2d(&markerRaise.x, &markerRaise.y, markerRotation);
 
+	Point shadowRaise;
+	shadowRaise.x = .5;
+	shadowRaise.y = -.75;
+	rotate2d(&shadowRaise.x, &shadowRaise.y, markerRotation);
+
+	ScreenMarker* marker = resetMarkerIterator();
+	searchMarkersCount = 0;
+	routeMarkersCount = 0;
+
+	while (marker != NULL) {
 		if (marker -> visible) {
+			renderedMarkers[count] = marker;
+			GLfloat tx = marker -> screenCoordinate.x - halfSize + markerRaise.x * marker -> raiseLevel / canvas.scale;
+			GLfloat ty = marker -> screenCoordinate.y - size + markerRaise.y * marker -> raiseLevel / canvas.scale;
 
-			GLfloat tx = marker -> screenCoordinate.x - searchMarkTexture.size / 2.0 / canvas.scale + normRaiseX * marker -> raiseLevel / canvas.scale;
-			GLfloat ty = marker -> screenCoordinate.y - searchMarkTexture.size / canvas.scale  + normRaiseY * marker -> raiseLevel / canvas.scale;
-			for (i = 0; i < VERTICES_PER_SEARCH_MARKER; i++) {
+			GLfloat tx2 = marker -> screenCoordinate.x - halfSize + shadowRaise.x * marker -> raiseLevel / canvas.scale;
+			GLfloat ty2 = marker -> screenCoordinate.y - size + shadowRaise.y * marker -> raiseLevel / canvas.scale;
+
+			for (i = 0; i < VERTICES_PER_QUAD; i++) {
+				int shadowIdx = verticeIdx;
 				searchMarkerVertices[verticeIdx++] = quadStripVertices[3 * i + 0] + tx;
 				searchMarkerVertices[verticeIdx++] = quadStripVertices[3 * i + 1] + ty;
 				searchMarkerVertices[verticeIdx++] = quadStripVertices[3 * i + 2];
+
+				shadowVertices[shadowIdx++] = quadStripVertices[3 * i + 0] + tx2;
+				shadowVertices[shadowIdx++] = quadStripVertices[3 * i + 1] + ty2;
+				shadowVertices[shadowIdx++] = quadStripVertices[3 * i + 2];
+
+			}
+
+			if (marker -> type == SEARCH_RESULT) {
+				searchMarkersCount++;
+			} else {
+				routeMarkersCount++;
 			}
 
 			count++;
-			if (count >= MAX_SEARCH_MARKERS - 1) {
+			if (count >= MAX_MARKERS - 1) {
 				break;
 			}
 		}
-		elem = elem -> next;
+		marker = markerIteratorNext();
 	}
-	memset(quadStripVertices, 0, 12 * sizeof(GLfloat));
+	resetQuadStripVertices();
 	return count;
 }
 
-void renderMarkers(int markerCount) {
-	glDisable(GL_TEXTURE_2D);
-	glEnable(GL_BLEND);
-
-//	glEnableClientState(GL_COLOR_ARRAY);
-//	glBindTexture(GL_TEXTURE_2D, 0);
-//	glColorPointer(4, GL_FLOAT, 0, searchMarkerShadowColors);
-//	glVertexPointer(3, GL_FLOAT, 0, searchMarkerShadowVertices);
-//	glDrawArrays(GL_TRIANGLES, 0, markerCount * VERTICES_PER_SEARCH_MARKER_SHADOW);
-//	glDisableClientState(GL_COLOR_ARRAY);
-
+void renderMarkerShadows(int count) {
 	glEnable(GL_TEXTURE_2D);
 	glBlendFunc(GL_DST_COLOR, GL_ZERO);
-	glTexCoordPointer(2, GL_FLOAT, 0, searchMarkerTextureCoords);
-	glVertexPointer(3, GL_FLOAT, 0, searchMarkerVertices);
+	glTexCoordPointer(2, GL_FLOAT, 0, quadTextureCoords);
+	glVertexPointer(3, GL_FLOAT, 0, shadowVertices);
 	glColor4f(1, 1, 1, 1);
-	glBindTexture(GL_TEXTURE_2D, searchMarkMask.name);
-	glDrawElements(GL_TRIANGLES, markerCount * VERTEX_INDICES_PER_SEARCH_MARKER, GL_UNSIGNED_SHORT, searchMarkerVertexIndices);
+	glBindTexture(GL_TEXTURE_2D, shadow.name);
+	glDrawElements(GL_TRIANGLES, count * VERTEX_INDICES_PER_QUAD, GL_UNSIGNED_SHORT, quadVertexIndices);
+}
+
+void renderMarkers(int start, int count, GLuint mask, GLuint texture) {
+	glEnable(GL_BLEND);
+	glBlendFunc(GL_DST_COLOR, GL_ZERO);
+	glTexCoordPointer(2, GL_FLOAT, 0, &quadTextureCoords[VERTICES_PER_QUAD * start * 2]);
+	glVertexPointer(3, GL_FLOAT, 0, &searchMarkerVertices[VERTEX_INDICES_PER_QUAD * start * 2]);
+	glColor4f(1, 1, 1, 1);
+	glBindTexture(GL_TEXTURE_2D, mask);
+	glDrawElements(GL_TRIANGLES, count * VERTEX_INDICES_PER_QUAD, GL_UNSIGNED_SHORT, quadVertexIndices);
 
 	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
-	glBindTexture(GL_TEXTURE_2D, searchMarkTexture.name);
-	glDrawElements(GL_TRIANGLES, markerCount * VERTEX_INDICES_PER_SEARCH_MARKER, GL_UNSIGNED_SHORT, searchMarkerVertexIndices);
+	glBindTexture(GL_TEXTURE_2D, texture);
+	glDrawElements(GL_TRIANGLES, count * VERTEX_INDICES_PER_QUAD, GL_UNSIGNED_SHORT, quadVertexIndices);
+}
+
+void drawDragCrosshair() {
+	GLfloat sizeX = dragCrosshair.size / canvas.scale;
+	GLfloat sizeY = dragCrosshair.size / canvas.scale;
+	setQuadStripSize(sizeX, sizeY, quadStripVertices);
+
+	double rotation = getMarkerRotation();
+	rotateMarkerStripVertices(rotation, sizeX / 2, sizeY / 2);
+
+	int verticeIdx = 0, i;
+
+	GLfloat tx = canvas.dragCrosshairMarker -> screenCoordinate.x - sizeX / 2.0;
+	GLfloat ty = canvas.dragCrosshairMarker -> screenCoordinate.y - sizeY / 2.0;
+
+	for (i = 0; i < VERTICES_PER_QUAD; i++) {
+		searchMarkerVertices[verticeIdx++] = quadStripVertices[3 * i + 0] + tx;
+		searchMarkerVertices[verticeIdx++] = quadStripVertices[3 * i + 1] + ty;
+		searchMarkerVertices[verticeIdx++] = quadStripVertices[3 * i + 2];
+	}
+	resetQuadStripVertices();
+	glBindTexture(GL_TEXTURE_2D, dragCrosshair.name);
+	glBlendFunc(GL_DST_COLOR, GL_ZERO);
+	glTexCoordPointer(2, GL_FLOAT, 0, quadTextureCoords);
+	glVertexPointer(3, GL_FLOAT, 0, searchMarkerVertices);
+	glDrawElements(GL_TRIANGLES, VERTEX_INDICES_PER_QUAD, GL_UNSIGNED_SHORT, quadVertexIndices);
+}
+
+void drawMouseHoldIndicator() {
+	static int mouseHoldIndicatorVerticesInitialized = FALSE;
+
+	if (mouseHoldIndicatorVerticesInitialized == FALSE) {
+		int i, verticeIdx = 0;
+		setQuadStripSize(12, 12, quadStripVertices);
+		for (i = 0; i < MAX_HOLD_INDICATOR_POINTS; i++) {
+			GLfloat tx = 0.0;
+			GLfloat ty = -40;
+			rotate2d(&tx, &ty, i * (360 / MAX_HOLD_INDICATOR_POINTS));
+			int j;
+			for (j = 0; j < VERTICES_PER_QUAD; j++) {
+				mouseHoldIndicatorVertices[verticeIdx++] = quadStripVertices[3 * j + 0] + tx - position -> mask.size / 2;
+				mouseHoldIndicatorVertices[verticeIdx++] = quadStripVertices[3 * j + 1] + ty - position -> mask.size / 2;
+				mouseHoldIndicatorVertices[verticeIdx++] = quadStripVertices[3 * j + 2];
+			}
+		}
+		mouseHoldIndicatorVerticesInitialized = TRUE;
+	}
+	if (canvas.mouseHoldInPlace > 0 && canvas.popup.popupEntries == NULL) {
+		int numberOfPoints = (canvas.mouseHoldInPlace / (double) options.mouseHoldToPopupTime) * (MAX_HOLD_INDICATOR_POINTS + 2);
+
+		if (numberOfPoints > MAX_HOLD_INDICATOR_POINTS) {
+			numberOfPoints = MAX_HOLD_INDICATOR_POINTS;
+		}
+
+		if (numberOfPoints > 0) {
+			glTranslatef(canvas.mouseHoldPositionSc.x, canvas.mouseHoldPositionSc.y, 0.0);
+			glEnable(GL_TEXTURE_2D);
+			glBlendFunc(GL_DST_COLOR, GL_ZERO);
+			glTexCoordPointer(2, GL_FLOAT, 0, quadTextureCoords);
+			glVertexPointer(3, GL_FLOAT, 0, mouseHoldIndicatorVertices);
+			glColor4f(1, 1, 1, 1);
+			glBindTexture(GL_TEXTURE_2D, position -> mask.name);
+			glScalef(1 / canvas.scale, 1 / canvas.scale, 1);
+			glDrawElements(GL_TRIANGLES, numberOfPoints * VERTEX_INDICES_PER_QUAD, GL_UNSIGNED_SHORT, quadVertexIndices);
+			glTranslatef(-canvas.mouseHoldPositionSc.x, -canvas.mouseHoldPositionSc.y, 0.0);
+		}
+	}
+}
+
+GLfloat getPopupEntryHeight() {
+	return 2 * options.popupEntryMargin + fontMediumMask.size / 16;
+}
+
+void drawPopup() {
+	if (canvas.popup.popupEntries == NULL) {
+		return;
+	}
+	glLoadIdentity();
+	glTranslatef(-SCREEN_WIDTH / 2 + (int) canvas.x - TILE_SIZE, -SCREEN_HEIGHT / 2 + (int) canvas.y - TILE_SIZE, -100);
+
+	if (canvas.popup.changed) {
+		canvas.popup.animationFrame = 0;
+		GList* elem = canvas.popup.popupEntries;
+		canvas.popup.height = 0;
+		canvas.popup.width = 0;
+		while (elem != NULL) {
+			PopupMenuEntry* entry = elem -> data;
+
+			canvas.popup.height += 2 * options.popupEntryMargin + fontMediumMask.size / 16;
+			entry -> displayNameWidth = stringWidth(entry -> displayName);
+
+			if (entry -> displayNameWidth > canvas.popup.width) {
+				canvas.popup.width = entry -> displayNameWidth;
+			}
+			elem = elem -> next;
+		}
+		canvas.popup.width += 2 * options.popupEntryMargin;
+		GLfloat x, y;
+		x = canvas.mouseHoldPositionSc.x - canvas.popup.width / 2;
+		y = canvas.mouseHoldPositionSc.y - canvas.popup.height / 2;
+		rotate2d(&x, &y, -canvas.rotz);
+		fprintf(stderr, "x = %f, y = %f\n", x, y);
+		canvas.popup.x = x;
+		canvas.popup.y = y;
+		canvas.popup.changed = FALSE;
+	}
+
+	glDisable(GL_TEXTURE_2D);
+	glTranslatef(canvas.popup.x, canvas.popup.y, 0);
+	glBlendFunc(GL_DST_COLOR, GL_ZERO);
+	if (canvas.popup.animationFrame < 30) {
+		canvas.popup.animationFrame++;
+		GLfloat scale = 0.5 + 0.5 * sin(canvas.popup.animationFrame / 25.0 * M_PI / 2) * 1.10;
+		glTranslatef(canvas.popup.width / 2.0 - canvas.popup.width / 2.0 * scale, canvas.popup.height / 2.0 - canvas.popup.height / 2.0 * scale, 0.0);
+		//		fprintf(stderr, "scale = %f\n", scale);
+		glScalef(scale, scale, scale);
+		glColor4f(.4, .4, .4, .7 * scale * scale);
+	} else {
+		glColor4f(.4, .4, .4, .7);
+	}
+
+	setQuadStripSize(canvas.popup.width, canvas.popup.height, quadStripVertices);
+	glVertexPointer(3, GL_FLOAT, 0, quadStripVertices);
+	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
+	glEnable(GL_TEXTURE_2D);
+
+	GLfloat popupEntryHeight = getPopupEntryHeight();
+
+	GList* elem = canvas.popup.popupEntries;
+	glTranslatef(canvas.popup.width / 2, options.popupEntryMargin, 0);
+	while (elem != NULL) {
+		PopupMenuEntry* entry = elem -> data;
+
+		if (entry == canvas.popup.clickedPopupEntry) {
+			glDisable(GL_TEXTURE_2D);
+			glTranslatef(-canvas.popup.width / 2, -options.popupEntryMargin, 0);
+			setQuadStripSize(canvas.popup.width, popupEntryHeight, quadStripVertices);
+			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
+			glColor4f(.0, .0, 1, .5);
+			glVertexPointer(3, GL_FLOAT, 0, quadStripVertices);
+			glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
+			glTranslatef(canvas.popup.width / 2, options.popupEntryMargin, 0);
+			glEnable(GL_TEXTURE_2D);
+		}
+
+		glTranslatef(-entry -> displayNameWidth / 2, 0, 0);
+		drawString(entry -> displayName, 1, 1, 1, FALSE);
+		glTranslatef(entry -> displayNameWidth / 2, popupEntryHeight, 0);
+		elem = elem -> next;
+	}
+
+	//	glEnableClientState(GL_COLOR_ARRAY);
+	//	line(0,0,50,50,10,0.9,.0,.2,.7,.0,.0,TRUE);
+	//	glDisableClientState(GL_COLOR_ARRAY);
 }
 
 void drawMarkers() {
-	if (canvas.markers == NULL) {
-		return;
+	prepareMarkers();
+	renderMarkerShadows(searchMarkersCount + routeMarkersCount);
+
+	if (searchMarkersCount > 0) {
+		renderMarkers(0, searchMarkersCount, searchMarkMask.name, searchMarkTexture.name);
 	}
-	int markerCount;
-	markerCount = prepareMarkers();
-	if (markerCount > 0) {
-		renderMarkers(markerCount);
+
+	int i = searchMarkersCount;
+	while (i < searchMarkersCount + routeMarkersCount) {
+		GLuint mask, texture;
+		if (renderedMarkers[i] -> type == ROUTE_START) {
+			mask = routeStartMask.name;
+			texture = routeStartTexture.name;
+		} else { //if (renderedMarkers[searchMarkersCount] -> type == ROUTE_START) {
+			mask = routeEndMask.name;
+			texture = routeEndTexture.name;
+		}
+		renderMarkers(i, 1, mask, texture);
+		i++;
+	}
+	if (canvas.draggedMarker != NULL || canvas.dragCrosshairMarker != NULL) {
+		if (canvas.draggedMarker != NULL) {
+			canvas.dragCrosshairMarker = canvas.draggedMarker;
+		}
+		if (canvas.dragCrosshairMarker -> raiseLevel == 0) {
+			canvas.dragCrosshairMarker = NULL;
+		} else {
+			drawDragCrosshair();
+		}
+	}
+}
+
+// http://answers.oreilly.com/topic/1669-how-to-render-anti-aliased-lines-with-textures-in-ios-4/
+int prepareRouteTriangleVertices() {
+	int i;
+	GArray* array = canvas.route.optimizedRoute.lineStripTileCoordinates;
+	if (array == NULL || array -> len < 2) {
+		return 0;
+	}
+	int idx = 0;
+
+	for (i = 0; i < array -> len - 1; i++) {
+		if (i >= MAX_ROUTE_SEGMENTS) {
+			return (i - 1);
+		}
+		TileCoordinate tc1 = g_array_index(array, TileCoordinate, i);
+		TileCoordinate tc2 = g_array_index(array, TileCoordinate, i + 1);
+		Point sc1;
+		Point sc2;
+		toScreenCoordinate(&tc1, &sc1);
+		toScreenCoordinate(&tc2, &sc2);
+		Point e; // east
+		e.x = sc2.x - sc1.x;
+		e.y = sc2.y - sc1.y;
+		GLfloat dist = pointDistance(&sc1, &sc2);
+		GLfloat scale;
+		if (dist == 0.0) {
+			e.x = 1.0;
+			e.y = 0.0;
+			scale = 1.0 / canvas.scale;
+		} else {
+			scale = 1.0 / dist / canvas.scale;
+		}
+		scale *= 4.0;
+
+		if (canvas.zoom > 14) {
+			scale *= 1.0 + (canvas.zoom - 16.0 + canvas.scale);
+		}
+		e.x *= scale;
+		e.y *= scale;
+
+		Point n; // north
+		n.x = -e.y;
+		n.y = e.x;
+
+		Point s; // south
+		s.x = -n.x;
+		s.y = -n.y;
+
+		Point w; // west
+		w.x = -e.x;
+		w.y = -e.y;
+
+		Point ne; // north-east
+		ne.x = n.x + e.x;
+		ne.y = n.y + e.y;
+
+		Point nw; // north-west
+		nw.x = n.x + w.x;
+		nw.y = n.y + w.y;
+
+		Point se; // south-east
+		se.x = s.x + e.x;
+		se.y = s.y + e.y;
+
+		Point sw; // south-west
+		sw.x = s.x + w.x;
+		sw.y = s.y + w.y;
+
+		if (i == 0) {
+
+		}
+
+		// starting half-circle
+		routeVertices[idx++] = sc1.x + sw.x;
+		routeVertices[idx++] = sc1.y + sw.y;
+		routeVertices[idx++] = 0.0;
+
+		routeVertices[idx++] = sc1.x + nw.x;
+		routeVertices[idx++] = sc1.y + nw.y;
+		routeVertices[idx++] = 0.0;
+
+		routeVertices[idx++] = sc1.x + s.x;
+		routeVertices[idx++] = sc1.y + s.y;
+		routeVertices[idx++] = 0.0;
+
+		routeVertices[idx++] = sc1.x + n.x;
+		routeVertices[idx++] = sc1.y + n.y;
+		routeVertices[idx++] = 0.0;
+
+		// main road segment
+		routeVertices[idx++] = sc1.x + s.x;
+		routeVertices[idx++] = sc1.y + s.y;
+		routeVertices[idx++] = 0.0;
+
+		routeVertices[idx++] = sc1.x + n.x;
+		routeVertices[idx++] = sc1.y + n.y;
+		routeVertices[idx++] = 0.0;
+
+		routeVertices[idx++] = sc2.x + s.x;
+		routeVertices[idx++] = sc2.y + s.y;
+		routeVertices[idx++] = 0.0;
+
+		routeVertices[idx++] = sc2.x + n.x;
+		routeVertices[idx++] = sc2.y + n.y;
+		routeVertices[idx++] = 0.0;
+
+		// ending half-circle
+		routeVertices[idx++] = sc2.x + s.x;
+		routeVertices[idx++] = sc2.y + s.y;
+		routeVertices[idx++] = 0.0;
+
+		routeVertices[idx++] = sc2.x + n.x;
+		routeVertices[idx++] = sc2.y + n.y;
+		routeVertices[idx++] = 0.0;
+
+		routeVertices[idx++] = sc2.x + se.x;
+		routeVertices[idx++] = sc2.y + se.y;
+		routeVertices[idx++] = 0.0;
+
+		routeVertices[idx++] = sc2.x + ne.x;
+		routeVertices[idx++] = sc2.y + ne.y;
+		routeVertices[idx++] = 0.0;
+
+	}
+
+	return (array -> len - 1);
+}
+
+void renderRouteTriangles(int count) {
+	glEnable(GL_BLEND);
+	glBlendFunc(GL_DST_COLOR, GL_ZERO);
+	glVertexPointer(3, GL_FLOAT, 0, routeVertices);
+	glBindTexture(GL_TEXTURE_2D, road.name);
+	glTexCoordPointer(2, GL_FLOAT, 0, routeTextCoords);
+	glColor4f(1.0, 1.0, 1.0, 1.0);
+	glDrawElements(GL_TRIANGLES, count * VERTEX_INDICES_PER_QUAD, GL_UNSIGNED_SHORT, quadVertexIndices);
+}
+
+void drawRoute() {
+	int routeSegments = prepareRouteTriangleVertices();
+	if (routeSegments > 0) {
+		renderRouteTriangles(routeSegments * QUADS_PER_ROUTE_SEGMENT);
 	}
 }
 
@@ -404,7 +795,7 @@ void drawBuildings() {
 	if (canvas.buildings == NULL) {
 		return;
 	}
-
+	glPushMatrix();
 	canvas.buildingsHeight = 20;
 	int i, j, k, verticeCount;
 	GLfloat translationX = (canvas.buildingsTileX - tiles[0][0][currentTilesIdx] -> tilex - getCanvasShiftX()) * TILE_SIZE;
@@ -437,45 +828,33 @@ void drawBuildings() {
 				}
 				verticeCount = wallCount * 18;
 				// 1st triangle, vertice 0
-				buildingsVertices[verticeCount++]
-				= g_array_index(polygon, double, j );
-				buildingsVertices[verticeCount++]
-				= g_array_index(polygon, double, j + 1);
+				buildingsVertices[verticeCount++] = g_array_index(polygon, double, j );
+				buildingsVertices[verticeCount++] = g_array_index(polygon, double, j + 1);
 				buildingsVertices[verticeCount++] = 0;
 
 				// 1st triangle, vertice 1
-				buildingsVertices[verticeCount++]
-				= g_array_index(polygon, double, j );
-				buildingsVertices[verticeCount++]
-				= g_array_index(polygon, double, j + 1);
+				buildingsVertices[verticeCount++] = g_array_index(polygon, double, j );
+				buildingsVertices[verticeCount++] = g_array_index(polygon, double, j + 1);
 				buildingsVertices[verticeCount++] = canvas.buildingsHeight;
 
 				// 1st triangle, vertice 2
-				buildingsVertices[verticeCount++]
-				= g_array_index(polygon, double, k);
-				buildingsVertices[verticeCount++]
-				= g_array_index(polygon, double, k + 1);
+				buildingsVertices[verticeCount++] = g_array_index(polygon, double, k);
+				buildingsVertices[verticeCount++] = g_array_index(polygon, double, k + 1);
 				buildingsVertices[verticeCount++] = 0;
 
 				// 2nd triangle, vertice 2
-				buildingsVertices[verticeCount++]
-				= g_array_index(polygon, double, k);
-				buildingsVertices[verticeCount++]
-				= g_array_index(polygon, double, k + 1);
+				buildingsVertices[verticeCount++] = g_array_index(polygon, double, k);
+				buildingsVertices[verticeCount++] = g_array_index(polygon, double, k + 1);
 				buildingsVertices[verticeCount++] = 0;
 
 				// 2nd triangle, vertice 1
-				buildingsVertices[verticeCount++]
-				= g_array_index(polygon, double, j );
-				buildingsVertices[verticeCount++]
-				= g_array_index(polygon, double, j + 1);
+				buildingsVertices[verticeCount++] = g_array_index(polygon, double, j );
+				buildingsVertices[verticeCount++] = g_array_index(polygon, double, j + 1);
 				buildingsVertices[verticeCount++] = canvas.buildingsHeight;
 
 				// 2nd triangle, vertice 3
-				buildingsVertices[verticeCount++]
-				= g_array_index(polygon, double, k);
-				buildingsVertices[verticeCount++]
-				= g_array_index(polygon, double, k + 1);
+				buildingsVertices[verticeCount++] = g_array_index(polygon, double, k);
+				buildingsVertices[verticeCount++] = g_array_index(polygon, double, k + 1);
 				buildingsVertices[verticeCount++] = canvas.buildingsHeight;
 
 				//            fprintf(stderr, "building triangle %d (i=%d/%d, j=%d/%d, k=%d)\n", wallCount, i, canvas.buildings -> len, j, polygon -> len/2, k);
@@ -498,7 +877,7 @@ void drawBuildings() {
 	glVertexPointer(3, GL_FLOAT, 0, buildingsVertices);
 	glDrawArrays(GL_TRIANGLES, 0, wallCount * 6);
 	glEnable(GL_TEXTURE_2D);
-	glTranslatef(-translationX, -translationY, 0.0);
+	glPopMatrix();
 }
 
 void drawTile_debug(t_tile *tile) {
@@ -577,12 +956,12 @@ void drawTile(t_tile* tile) {
 	}
 
 	// TODO tmp
-//	if ((tile -> state == STATE_LOADED) && nowMillis - tile->stateChangeTime < tile->transitionTime) {
-//		color = (nowMillis - tile->stateChangeTime) / (float) tile->transitionTime;
-		//		color = 1.0 - sqrt(sqrt(1.0 - color));
-//	} else {
-		color = 1.0;
-//	}
+	//	if ((tile -> state == STATE_LOADED) && nowMillis - tile->stateChangeTime < tile->transitionTime) {
+	//		color = (nowMillis - tile->stateChangeTime) / (float) tile->transitionTime;
+	//		color = 1.0 - sqrt(sqrt(1.0 - color));
+	//	} else {
+	color = 1.0;
+	//	}
 
 	// If there is old tile to display and loaded tile to blend with it use multitexturing for smoother transition.
 	if (tile -> state == STATE_LOADED && color < 1.0 && tile -> oldTiles[0] != NULL) {
@@ -769,28 +1148,36 @@ SDL_GLES_Context *context;
 
 /* general OpenGL initialization function */
 int initGL(GLvoid) {
+
 	SDL_Surface *surface;
-	if (SDL_Init(SDL_INIT_EVERYTHING | SDL_INIT_NOPARACHUTE) < 0) {
+	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE) < 0) {
 		fprintf(stderr, "SDL initialization failed: %s\n", SDL_GetError());
 		exit(1);
 	}
 
-	if (TTF_Init() == -1) {
-		fprintf(stderr, "TTF_Init: %s\n", TTF_GetError());
-		exit(2);
-	}
+//	if (TTF_Init() == -1) {
+//		fprintf(stderr, "TTF_Init: %s\n", TTF_GetError());
+//		exit(2);
+//	}
 
-	surface = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 16,
-#ifdef N900
-			SDL_FULLSCREEN
+#if defined(N900)
+	surface = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 16, SDL_FULLSCREEN);
+#elif defined(N950)
+	fprintf(stderr, "N950 opengl init\n");
+	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 1);
+	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
+	surface = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 16, SDL_OPENGLES | SDL_FULLSCREEN);
 #else
-			SDL_OPENGL
+	surface = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 16, SDL_OPENGL);
 #endif
-			);
+	if (surface == NULL) {
+		fprintf(stderr, "Couldn't set GL mode: %s\n", SDL_GetError());
+		exit(-1);
+	}
 	SDL_ShowCursor(0);
 
 	SDL_WM_SetCaption("CloudGPS", "cloudgps");
-#ifdef N900
+#if defined(N900)
 	SDL_GLES_Init(SDL_GLES_VERSION_1_1);
 
 	context = SDL_GLES_CreateContext();
@@ -850,7 +1237,7 @@ int initGL(GLvoid) {
 	glViewport(0, 0, (GLint) SCREEN_WIDTH, (GLint) SCREEN_HEIGHT);
 	glMatrixMode(GL_PROJECTION);
 	glLoadIdentity();
-#ifdef N900
+#if defined(N900) || defined(N950)
 	glFrustumf(-(SCREEN_WIDTH/4), (SCREEN_WIDTH/4), (SCREEN_HEIGHT/4), -(SCREEN_HEIGHT/4), 50.0, 600);
 #else
 	//	GLfloat left, right, bottom, top, near=20, far=300;
@@ -862,7 +1249,7 @@ int initGL(GLvoid) {
 	//	left = aspect * top;
 
 	//	glFrustum(left, right, bottom, top, near, far);
-	glFrustum(-(SCREEN_WIDTH / 4), (SCREEN_WIDTH / 4), (SCREEN_HEIGHT / 4), -(SCREEN_HEIGHT / 4), 50.0, 300);
+	glFrustum(-(SCREEN_WIDTH / 4), (SCREEN_WIDTH / 4), (SCREEN_HEIGHT / 4), -(SCREEN_HEIGHT / 4), 50.0, 600);
 #endif
 	glMatrixMode(GL_MODELVIEW);
 	glLoadIdentity();
@@ -879,6 +1266,9 @@ int initGL(GLvoid) {
 	//    const GLubyte *extensions = glGetString(GL_EXTENSIONS);
 	//    fprintf(stderr, "GL extensions: %s\n", (const char*) extensions);
 
+	initQuadIndices(quadVertexIndices, MAX_QUADS);
+	initQuadTextureCoords();
+	initRoadTextureCoords();
 	return (TRUE);
 }
 
@@ -927,110 +1317,8 @@ void setupGlScene() {
 	storeMatricesAndViewPort();
 }
 
-void drawPosArrow() {
-	GLfloat arrow[9];
-	arrow[0] = canvas.arrowPosX;
-	arrow[1] = canvas.arrowPosY;
-	arrow[2] = 0.0;
-	arrow[5] = 0.0;
-	arrow[8] = 0.0;
-	int size = 20;
-
-	switch (canvas.arrowPos) {
-	case 1: // left
-		arrow[3] = canvas.arrowPosX + size;
-		arrow[4] = canvas.arrowPosY - size / 2;
-
-		arrow[6] = canvas.arrowPosX + size;
-		arrow[7] = canvas.arrowPosY + size / 2;
-		break;
-	case 2: // right
-		arrow[3] = canvas.arrowPosX - size;
-		arrow[4] = canvas.arrowPosY - size / 2;
-
-		arrow[6] = canvas.arrowPosX - size;
-		arrow[7] = canvas.arrowPosY + size / 2;
-		break;
-	case 3: // up
-		arrow[3] = canvas.arrowPosX - size / 2;
-		arrow[4] = canvas.arrowPosY + size;
-
-		arrow[6] = canvas.arrowPosX + size / 2;
-		arrow[7] = canvas.arrowPosY + size;
-		break;
-	case 4: // down
-		arrow[3] = canvas.arrowPosX - size / 2;
-		arrow[4] = canvas.arrowPosY - size;
-
-		arrow[6] = canvas.arrowPosX + size / 2;
-		arrow[7] = canvas.arrowPosY - size;
-		break;
-	}
-
-	glColor4f(.0, .0, 1.0, 0.6);
-	glVertexPointer(3, GL_FLOAT, 0, &arrow);
-	glDrawArrays(GL_TRIANGLES, 0, 3);
-}
-
-void drawGLScene() {
-	static GLint oldTicks = 0;
-	static GLint frames = 0;
+void drawDebugTiles() {
 	int i, j;
-	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
-
-	GLenum err = glGetError();
-	if (err != GL_NO_ERROR) {
-		fprintf(stderr, "627 GL_ERROR: %x\n", err);
-	}
-
-	for (i = 0; i < TILES_X; i++) {
-		for (j = 0; j < TILES_Y; j++) {
-			if (tiles[i][j][currentTilesIdx] != NULL) {
-				if (tiles[i][j][currentTilesIdx] -> visible == TRUE) {
-					glTranslatef((i - getCanvasShiftX()) * TILE_SIZE, (j - getCanvasShiftY()) * TILE_SIZE, 0);
-					drawTile(tiles[i][j][currentTilesIdx]);
-					glTranslatef(-(i - getCanvasShiftX()) * TILE_SIZE, -(j - getCanvasShiftY()) * TILE_SIZE, 0);
-				}
-			}
-		}
-	}
-
-	glActiveTexture(GL_TEXTURE1);
-	glDisable(GL_TEXTURE_2D);
-	glActiveTexture(GL_TEXTURE0);
-	glClientActiveTexture(GL_TEXTURE0);
-	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
-
-	drawBuildings();
-
-	if (position -> status != UI_HIDDEN) {
-		// Current position
-		drawUiElement(position);
-
-		// Arrow showing direction of movement
-#ifdef N900
-		if ((device -> fix -> fields | LOCATION_GPS_DEVICE_TRACK_SET) && device -> satellites_in_use > 0 && device -> fix -> speed > 1) {
-			GLfloat track = device -> fix -> track - 180.0;
-			glTranslatef(position -> x + position -> texture.size / 2 / canvas.scale, position -> y + position -> texture.size / 2 / canvas.scale,
-					0.0);
-			glDisable(GL_TEXTURE_2D);
-			glColor4f(0.09, 0.59, .94, 1.0);
-			glRotatef(track, 0.0, 0.0, 1.0);
-			directionVertices[1] = 18.0f / canvas.scale;
-			directionVertices[3] = -6.0f / canvas.scale;
-			directionVertices[4] = 6.0f / canvas.scale;
-			directionVertices[6] = 6.0f / canvas.scale;
-			directionVertices[7] = 6.0f / canvas.scale;
-
-			glVertexPointer(3, GL_FLOAT, 0, directionVertices);
-			glDrawArrays(GL_TRIANGLES, 0, 3);
-			glRotatef(-track, 0.0, 0.0, 1.0);
-			glTranslatef(-position -> x - position -> texture.size / 2, -position -> y - position -> texture.size / 2, 0.0);
-		}
-#endif
-	}
-
-	drawMarkers();
 	// Small grid of all tiles coloured by their state. Also a bounding trapezoid
 	if (options.showDebugTiles) {
 		glLoadIdentity();
@@ -1119,13 +1407,60 @@ void drawGLScene() {
 		glScalef(20, 20, 20);
 		glTranslatef(-150, -20, 0);
 		glEnable(GL_TEXTURE_2D);
-		sprintf(strbuf, "intZoom:%d, destZoom:%d, zoom:%d", canvas.attractionIntermediateZoom, canvas.attractionDestZoom, canvas.zoom);
-		drawString(strbuf, .2, .0, 0.5, TRUE);
+		//		sprintf(strbuf, "routingStatus = %d", routingStatus);
+		//		drawString(strbuf, .2, .0, 0.5, TRUE);
 		glTranslatef(0, -20, 0);
-		sprintf(strbuf, "allocatedTiles: %d, dq: %d", allCreatedTiles -> length, downloadQueue -> length);
+		sprintf(strbuf, "minIdx: %d, flybyIdx: %d", canvas.navigationStatus.currentLineIdx, canvas.routeFlybyLineIdx);
 		drawString(strbuf, .2, .0, 0.5, TRUE);
 	}
+}
 
+void drawPosArrow() {
+	GLfloat arrow[9];
+	arrow[0] = canvas.arrowPosX;
+	arrow[1] = canvas.arrowPosY;
+	arrow[2] = 0.0;
+	arrow[5] = 0.0;
+	arrow[8] = 0.0;
+	int size = 20;
+
+	switch (canvas.arrowPos) {
+	case 1: // left
+		arrow[3] = canvas.arrowPosX + size;
+		arrow[4] = canvas.arrowPosY - size / 2;
+
+		arrow[6] = canvas.arrowPosX + size;
+		arrow[7] = canvas.arrowPosY + size / 2;
+		break;
+	case 2: // right
+		arrow[3] = canvas.arrowPosX - size;
+		arrow[4] = canvas.arrowPosY - size / 2;
+
+		arrow[6] = canvas.arrowPosX - size;
+		arrow[7] = canvas.arrowPosY + size / 2;
+		break;
+	case 3: // up
+		arrow[3] = canvas.arrowPosX - size / 2;
+		arrow[4] = canvas.arrowPosY + size;
+
+		arrow[6] = canvas.arrowPosX + size / 2;
+		arrow[7] = canvas.arrowPosY + size;
+		break;
+	case 4: // down
+		arrow[3] = canvas.arrowPosX - size / 2;
+		arrow[4] = canvas.arrowPosY - size;
+
+		arrow[6] = canvas.arrowPosX + size / 2;
+		arrow[7] = canvas.arrowPosY - size;
+		break;
+	}
+
+	glColor4f(.0, .0, 1.0, 0.6);
+	glVertexPointer(3, GL_FLOAT, 0, &arrow);
+	glDrawArrays(GL_TRIANGLES, 0, 3);
+}
+
+void drawFog() {
 	glLoadIdentity();
 	glTranslatef(-SCREEN_WIDTH / 2, -SCREEN_HEIGHT / 2, -100);
 
@@ -1170,7 +1505,124 @@ void drawGLScene() {
 		glDisableClientState(GL_COLOR_ARRAY);
 		glTranslatef(fogStart, 0, 0);
 	}
+}
 
+void drawCoordinates() {
+	glTranslatef(7, 7, 0);
+	glColor4f(0.1, 0.1, 0.1, 0.6);
+	sprintf(strbuf, "%04.6f %04.6f", canvas.center.latitude, canvas.center.longitude);
+
+	setQuadStripSize(stringWidth(strbuf) + 10, 40, quadStripVertices);
+	glVertexPointer(3, GL_FLOAT, 0, quadStripVertices);
+	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
+
+	glTranslatef(7, 5, 0);
+	drawString(strbuf, 1.0, 1.0, 1.0, FALSE);
+	glTranslatef(0, 15, 0);
+	double distance = getDistance(&canvas.center, &canvas.currentPos);
+
+	if (distance < 1000) {
+		sprintf(strbuf, "distance = %.0fm", distance);
+	} else if (distance < 100000) {
+		sprintf(strbuf, "distance = %.2fkm", distance / 1000);
+	} else {
+		sprintf(strbuf, "distance = %.0fkm", distance / 1000);
+	}
+	drawString(strbuf, 1.0, 1.0, 1.0, FALSE);
+}
+
+void drawMovementDirectionArrow() {
+	// Arrow showing direction of movement
+#ifdef N900
+	if ((device -> fix -> fields | LOCATION_GPS_DEVICE_TRACK_SET) && device -> satellites_in_use > 0 && device -> fix -> speed > 1) {
+		GLfloat track = device -> fix -> track - 180.0;
+		glTranslatef(position -> x + position -> texture.size / 2 / canvas.scale, position -> y + position -> texture.size / 2 / canvas.scale,
+				0.0);
+		glDisable(GL_TEXTURE_2D);
+		glColor4f(0.09, 0.59, .94, 1.0);
+		glRotatef(track, 0.0, 0.0, 1.0);
+		directionVertices[1] = 18.0f / canvas.scale;
+		directionVertices[3] = -6.0f / canvas.scale;
+		directionVertices[4] = 6.0f / canvas.scale;
+		directionVertices[6] = 6.0f / canvas.scale;
+		directionVertices[7] = 6.0f / canvas.scale;
+
+		glVertexPointer(3, GL_FLOAT, 0, directionVertices);
+		glDrawArrays(GL_TRIANGLES, 0, 3);
+		glRotatef(-track, 0.0, 0.0, 1.0);
+		glTranslatef(-position -> x - position -> texture.size / 2, -position -> y - position -> texture.size / 2, 0.0);
+	}
+#endif
+}
+
+void computeFps() {
+	static GLint oldTicks = 0;
+	static GLint frames = 0;
+	frames++;
+	{
+		GLint t = SDL_GetTicks();
+		if (t - oldTicks >= 1000) {
+			seconds = (t - oldTicks) / 1000.0;
+			fps = frames / seconds;
+			oldTicks = t;
+			frames = 0;
+			//            fprintf(stderr, "fps: %f\n", fps);
+		}
+	}
+}
+
+void swapBuffers() {
+#ifdef N900
+	SDL_GLES_SwapBuffers();
+#else
+	SDL_GL_SwapBuffers();
+#endif
+}
+
+void drawTiles() {
+	int i, j;
+	for (i = 0; i < TILES_X; i++) {
+		for (j = 0; j < TILES_Y; j++) {
+			if (tiles[i][j][currentTilesIdx] != NULL) {
+				if (tiles[i][j][currentTilesIdx] -> visible == TRUE) {
+					glTranslatef((i - getCanvasShiftX()) * TILE_SIZE, (j - getCanvasShiftY()) * TILE_SIZE, 0);
+					drawTile(tiles[i][j][currentTilesIdx]);
+					glTranslatef(-(i - getCanvasShiftX()) * TILE_SIZE, -(j - getCanvasShiftY()) * TILE_SIZE, 0);
+				}
+			}
+		}
+	}
+
+	glActiveTexture(GL_TEXTURE1);
+	glDisable(GL_TEXTURE_2D);
+	glActiveTexture(GL_TEXTURE0);
+	glClientActiveTexture(GL_TEXTURE0);
+	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
+}
+
+void drawGLScene() {
+	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
+
+	GLenum err = glGetError();
+	if (err != GL_NO_ERROR) {
+		fprintf(stderr, "drawGLScene GL_ERROR: %x\n", err);
+	}
+
+	drawTiles();
+	drawBuildings();
+
+	if (position -> status != UI_HIDDEN) {
+		// Current position
+		drawUiElement(position);
+		drawMovementDirectionArrow();
+	}
+
+	drawRoute();
+	drawMarkers();
+	drawMouseHoldIndicator();
+	drawPopup();
+	drawDebugTiles();
+	drawFog();
 	drawUiElems();
 	drawZoomBar();
 	drawSearchBar();
@@ -1186,43 +1638,9 @@ void drawGLScene() {
 		drawPosArrow();
 	}
 	if (options.showCoordinates) {
-		glTranslatef(7, 7, 0);
-		glColor4f(0.1, 0.1, 0.1, 0.6);
-		sprintf(strbuf, "%04.6f %04.6f", canvas.center.latitude, canvas.center.longitude);
-
-		setQuadStripSize(stringWidth(strbuf) + 10, 40, quadStripVertices);
-		glVertexPointer(3, GL_FLOAT, 0, quadStripVertices);
-		glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
-
-		glTranslatef(7, 5, 0);
-		drawString(strbuf, 1.0, 1.0, 1.0, FALSE);
-		glTranslatef(0, 15, 0);
-		double distance = getDistance(&canvas.center, &canvas.currentPos);
-
-		if (distance < 1000) {
-			sprintf(strbuf, "distance = %.0fm", distance);
-		} else if (distance < 100000) {
-			sprintf(strbuf, "distance = %.2fkm", distance / 1000);
-		} else {
-			sprintf(strbuf, "distance = %.0fkm", distance / 1000);
-		}
-		drawString(strbuf, 1.0, 1.0, 1.0, FALSE);
+		drawCoordinates();
 	}
-
-	frames++;
-	{
-		GLint t = SDL_GetTicks();
-		if (t - oldTicks >= 1000) {
-			seconds = (t - oldTicks) / 1000.0;
-			fps = frames / seconds;
-			oldTicks = t;
-			frames = 0;
-			//            fprintf(stderr, "fps: %f\n", fps);
-		}
-	}
-#ifdef N900
-	SDL_GLES_SwapBuffers();
-#else
-	SDL_GL_SwapBuffers();
-#endif
+	computeFps();
+	swapBuffers();
 }
+
