@@ -22,6 +22,9 @@
 #include "glunproject.h"
 #include "downloadQueue.h"
 #include "geocoding/geocoder.h"
+#include "routing/router.h"
+#include "markerIterator.h"
+#include "geometry.h"
 
 #ifdef N900
 #define DOWNLOADING_THREADS 2
@@ -32,6 +35,7 @@ extern Canvas canvas;
 UiElement* pressedUiElem;
 
 SDL_Thread* downloadThreads[DOWNLOADING_THREADS];
+volatile int downloadThreadBusy[DOWNLOADING_THREADS];
 
 int downloadThread(void* queue);
 void tileEngine_computeBoundingTrapezoid();
@@ -70,6 +74,15 @@ void toScreenCoordinate(TileCoordinate *tc, Point* p) {
 	p -> x = (tc->tilex - canvas.tilex - getCanvasShiftX()) * TILE_SIZE + tc->x;
 	p -> y = (tc->tiley - canvas.tiley - getCanvasShiftY()) * TILE_SIZE + tc->y;
 }
+int adjustTileCoordinateShifts(TileCoordinate* tc);
+// Warning: does not handle zoom
+void fromScreenCoordinate(Point* p, TileCoordinate *tc) {
+	tc -> tilex = 0;
+	tc -> tiley = 0;
+	tc -> x = p -> x + (canvas.tilex + getCanvasShiftX()) * TILE_SIZE;
+	tc -> y = p -> y + (canvas.tiley + getCanvasShiftY()) * TILE_SIZE;
+	adjustTileCoordinateShifts(tc);
+}
 
 /* Ground resolution (in meters per pixel) for given tile coordinate. */
 double getGroundResolution(WorldCoordinate* c, int zoom) {
@@ -194,6 +207,9 @@ void createZoomTile(TileCoordinate* tc, int zoomTileIdx, int zoom) {
 			}
 		}
 
+		if (tc -> zoom > canvas.zoom + 3) {
+			newTile = FALSE;
+		}
 		if (tc -> zoom < canvas.provider -> minZoom || tc -> zoom > canvas.provider -> maxZoom) {
 			newTile = FALSE;
 		}
@@ -274,10 +290,10 @@ void tileEngine_handleZoomTiles() {
 					int k, l = 1;
 					k = tiles[i][j][currentTilesIdx] -> zoom;
 
-					if (canvas.destScale == 0.0 && canvas.scale != 1.0) {
-						k = tiles[i][j][currentTilesIdx] -> zoom - 1;
-						l = 2;
-					}
+					//if (canvas.destScale == 0.0 && canvas.scale != 1.0) {
+					k = tiles[i][j][currentTilesIdx] -> zoom - 1;
+					l = 2;
+					//}
 
 					if (k <= 0) {
 						k = 1;
@@ -353,9 +369,9 @@ void initTileEngine() {
 
 	for (i = 0; i < DOWNLOADING_THREADS; i++) {
 		//        fprintf(stderr, "before SDL_CreateThread\n");
-		downloadThreads[i] = SDL_CreateThread(downloadThread, (void*) downloadQueue);
+		downloadThreads[i] = SDL_CreateThread(downloadThread, (void*) &downloadThreadBusy[i]);
 		//        fprintf(stderr, "after SDL_CreateThread\n");
-//		SDL_Delay(15);
+		//		SDL_Delay(15);
 	}
 
 	for (i = 0; i < TILES_X; i++) {
@@ -391,24 +407,24 @@ void deallocateSearchMarkers() {
 	}
 }
 
-void addGeoCodeResult(GeoCodeResult* geoCodeResult) {
-	if (geoCodeResult -> searchId == canvas.lastSearchId) {
-		ScreenMarker* marker = calloc(1, sizeof(ScreenMarker));
+ScreenMarker* createMarker(GeoCodeResult* geoCodeResult, MarkerType type) {
+	ScreenMarker* marker = calloc(1, sizeof(ScreenMarker));
+	if (marker != NULL) {
+		marker -> type = type;
 		marker -> tileCoordinate.zoom = -1;
 		marker -> visible = TRUE;
 		marker -> raised = FALSE;
 		marker -> raiseLevel = 50;
 		marker -> geoCodeResult = geoCodeResult;
-		canvas.markers = g_list_prepend(canvas.markers, marker);
-	} else {
-		free(geoCodeResult);
 	}
+	return marker;
 }
 
 void addSearchMarkers(GList* searchResults) {
 	while (searchResults != NULL) {
 		GeoCodeResult* geoCodeResult = (GeoCodeResult*) searchResults -> data;
-		addGeoCodeResult(geoCodeResult);
+		ScreenMarker* marker = createMarker(geoCodeResult, SEARCH_RESULT);
+		canvas.markers = g_list_prepend(canvas.markers, marker);
 		searchResults = searchResults -> next;
 	}
 }
@@ -484,9 +500,129 @@ void tileEngineProcessSearchResults(GList* searchResults) {
 	setAttractionToSearchResults();
 }
 
+void tileEngineFreeRoute() {
+	if (canvas.route.directions != NULL) {
+		g_array_free(canvas.route.directions, TRUE);
+		canvas.route.directions = NULL;
+	}
+	if (canvas.route.optimizedRoute.lineStripTileCoordinates != NULL) {
+		g_array_free(canvas.route.optimizedRoute.lineStripTileCoordinates, TRUE);
+		canvas.route.optimizedRoute.lineStripTileCoordinates = NULL;
+	}
+	if (canvas.route.optimizedRoute.visibleLines != NULL) {
+		g_list_free(canvas.route.optimizedRoute.visibleLines);
+		canvas.route.optimizedRoute.visibleLines = NULL;
+	}
+	if (canvas.route.query != NULL) {
+		if (canvas.route.query -> waypoints != NULL) {
+			g_list_free(canvas.route.query -> waypoints);
+			canvas.route.query -> waypoints = NULL;
+		}
+		free(canvas.route.query);
+		canvas.route.query = NULL;
+	}
+	canvas.route.lengthMeters = 0;
+}
+
+void tileEngineUpdateOptimizedRoute() {
+	if (canvas.route.optimizedRoute.lineStripZoom != canvas.zoom) {
+		canvas.route.optimizedRoute.lineStripTileCoordinates = NULL;
+	}
+	if (canvas.route.directions == NULL) {
+		return;
+	}
+
+	if (canvas.route.optimizedRoute.lineStripTileCoordinates == NULL) {
+		int i, j;
+		canvas.route.optimizedRoute.lineStripTileCoordinates = g_array_new(FALSE, FALSE, sizeof(TileCoordinate));
+		if (canvas.route.optimizedRoute.lineStripTileCoordinates == NULL) {
+			fprintf(stderr, "Memory allocation error in tileEngineRecalculateOptimizedRoute\n");
+			exit(1);
+		}
+		for (i = 0; i < canvas.route.directions -> len; i++) {
+			RouteDirection direction = g_array_index(canvas.route.directions, RouteDirection, i);
+			for (j = 0; j < direction.polyLine -> len; j++) {
+				WorldCoordinate wc = g_array_index(direction.polyLine, WorldCoordinate, j);
+				TileCoordinate tc;
+				toTileCoordinate(&wc, &tc, canvas.zoom);
+				canvas.route.optimizedRoute.lineStripTileCoordinates = g_array_append_vals(canvas.route.optimizedRoute.lineStripTileCoordinates, &tc, 1);
+			}
+		}
+		canvas.route.optimizedRoute.lineStripZoom = canvas.zoom;
+	}
+}
+
+GeoCodeResult* createGeoCodeResult(WorldCoordinate* wc) {
+	GeoCodeResult* r = calloc(1, sizeof(GeoCodeResult));
+	r -> centroid = *wc;
+	return r;
+}
+
+void addRouteMarker(WorldCoordinate* wc, MarkerType type) {
+	GeoCodeResult* geoCode = createGeoCodeResult(wc);
+	ScreenMarker* marker = createMarker(geoCode, type);
+	canvas.routeMarkers = g_list_append(canvas.routeMarkers, marker);
+}
+
+void clearRouteMarkers() {
+	if (canvas.routeMarkers != NULL) {
+		g_list_free(canvas.routeMarkers);
+		canvas.routeMarkers = NULL;
+	}
+}
+
+void createRouteMarkers() {
+	addRouteMarker(&canvas.route.query -> start, ROUTE_START);
+	GList* waypoint = canvas.route.query -> waypoints;
+
+	while (waypoint != NULL) {
+		WorldCoordinate* wc = (WorldCoordinate*) waypoint -> data;
+		addRouteMarker(wc, ROUTE_WAYPOINT);
+		waypoint = waypoint -> next;
+	}
+	addRouteMarker(&canvas.route.query -> end, ROUTE_END);
+}
+
+void updateMarkerTileCoordinate(ScreenMarker* marker);
+void updateMarkerScreenCoordinate(ScreenMarker* marker);
+void updateRouteMarker(ScreenMarker* marker, WorldCoordinate* wc) {
+	marker -> geoCodeResult -> centroid = *wc;
+	marker -> recalculateCoords = TRUE;
+	marker -> raised = 0;
+	marker -> raiseLevel = 50;
+	updateMarkerTileCoordinate(marker);
+	updateMarkerScreenCoordinate(marker);
+}
+
+void updateRouteMarkers(MarkerType type) {
+	GList* elem = canvas.routeMarkers;
+
+	while (elem != NULL) {
+		ScreenMarker* marker = elem -> data;
+		if (marker -> type == type) {
+			if (marker -> type == ROUTE_START) {
+				updateRouteMarker(marker, &canvas.route.query -> start);
+			} else if (marker -> type == ROUTE_END) {
+				updateRouteMarker(marker, &canvas.route.query -> end);
+			}
+		}
+		elem = elem -> next;
+	}
+}
+
+void tileEngineProcessNewRoute(Route* route) {
+	tileEngineFreeRoute();
+	canvas.route = *route;
+	memset(route, 0, sizeof(Route));
+	if (canvas.routeMarkers == NULL) {
+		createRouteMarkers();
+	}
+}
+
 void updateMarkerTileCoordinate(ScreenMarker* marker) {
-	if (marker -> tileCoordinate.zoom != canvas.zoom) {
+	if (marker -> tileCoordinate.zoom != canvas.zoom || marker -> recalculateCoords) {
 		toTileCoordinate(&marker -> geoCodeResult -> centroid, &marker -> tileCoordinate, canvas.zoom);
+		marker -> recalculateCoords = FALSE;
 	}
 }
 
@@ -518,8 +654,7 @@ void updateMarkerVisibility(ScreenMarker* marker) {
 	}
 }
 
-void tileEngineUpdateMarkers() {
-	GList* list = canvas.markers;
+void tileEngineUpdateMarkerList(GList* list) {
 	while (list != NULL) {
 		ScreenMarker* marker = (ScreenMarker*) list -> data;
 		updateMarkerTileCoordinate(marker);
@@ -530,6 +665,30 @@ void tileEngineUpdateMarkers() {
 	}
 }
 
+void updateBusyValue() {
+	int i;
+	canvas.busy = FALSE;
+
+	for (i = 0; i < DOWNLOADING_THREADS; i++) {
+		if (downloadThreadBusy[i]) {
+			canvas.busy = TRUE;
+		}
+	}
+
+	if (routingStatus != NO_QUERY) {
+		canvas.busy = TRUE;
+	}
+	if (searchResultsStatus != NO_QUERY) {
+		canvas.busy = TRUE;
+
+	}
+}
+
+void tileEngineUpdateMarkers() {
+	tileEngineUpdateMarkerList(canvas.markers);
+	tileEngineUpdateMarkerList(canvas.routeMarkers);
+}
+
 int adjustTileCoordinateShifts(TileCoordinate* tc) {
 	int horizontalShift = roundDown(tc -> x / (double) TILE_SIZE);
 	int verticalShift = roundDown(tc -> y / (double) TILE_SIZE);
@@ -770,84 +929,307 @@ void setClickedTileCoordinate(TileCoordinate* tc) {
 	adjustTileCoordinateShifts(tc);
 }
 
-int isInsideBox(Point* p, GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2) {
-	if (p -> x >= x1 && p -> x <= x2 && p -> y >= y1 && p -> y <= y2) {
-		return TRUE;
-	}
-	return FALSE;
-}
-
 GLfloat getMarkerRotation() {
-	return -canvas.rotz - pow(canvas.orientationTransitionLinear/ 160.0, 1.5) * 90.0;
+	return -canvas.rotz - pow(canvas.orientationTransitionLinear / 160.0, 1.5) * 90.0;
 }
 
 extern Texture searchMarkTexture;
 void rotate2d(GLfloat* x, GLfloat* y, double rotation);
-void checkClickedMarkers() {
+
+ScreenMarker* findMarkerUnderMouse() {
 	TileCoordinate tc;
 	Point clicked;
 
 	setClickedTileCoordinate(&tc);
 	toScreenCoordinate(&tc, &clicked);
-	GLfloat x1, y1;
-	x1 = 0.0;
-	y1 = searchMarkTexture.size / 2;
-	rotate2d(&x1, &y1, getMarkerRotation());
+	GLfloat textureX, textureY;
+	textureX = 0.0;
+	textureY = searchMarkTexture.size / 2;
+	rotate2d(&textureX, &textureY, getMarkerRotation());
 
-	x1 += clicked.x;
-	y1 += clicked.y;
+	Point screenCoordinateClicked;
 
-	GList* list = canvas.markers;
-	while (list != NULL) {
-		ScreenMarker* marker = (ScreenMarker*) list -> data;
+	screenCoordinateClicked.x = textureX + clicked.x;
+	screenCoordinateClicked.y = textureY + clicked.y;
+
+	ScreenMarker* closestMarker = NULL;
+
+	ScreenMarker* marker = resetMarkerIterator();
+	GLfloat minDistance = searchMarkTexture.size / 2;
+	while (marker != NULL) {
 		if (marker -> visible) {
-			if (isInsideBox(&marker -> screenCoordinate, x1 - searchMarkTexture.size / 2, y1 - searchMarkTexture.size / 2, x1 + searchMarkTexture.size / 2, y1 + searchMarkTexture.size / 2)) {
-				setAttractionToPoint(&marker -> geoCodeResult -> centroid);
-				if (marker -> geoCodeResult -> description != NULL) {
-					addConsoleLine((char *) marker -> geoCodeResult -> description, 0, 1, 1);
-				}
-				break;
+			GLfloat distance = pointDistance(&marker -> screenCoordinate, &screenCoordinateClicked);
+
+			if (distance < minDistance) {
+				minDistance = distance;
+				closestMarker = marker;
 			}
 		}
-		list = list -> next;
+		marker = markerIteratorNext();
 	}
-}
-
-void addMarkerOnClickedPosition() {
-	TileCoordinate tc;
-
-	double x2 = (mouse.oldx - SCREEN_WIDTH / 2) * cos(-canvas.rotz * M_PI / 180.0) - (mouse.oldy - SCREEN_HEIGHT / 2) * sin(-canvas.rotz * M_PI / 180.0);
-	double y2 = (mouse.oldx - SCREEN_WIDTH / 2) * sin(-canvas.rotz * M_PI / 180.0) + (mouse.oldy - SCREEN_HEIGHT / 2) * cos(-canvas.rotz * M_PI / 180.0);
-	x2 /= canvas.scale;
-	y2 /= canvas.scale;
-
-	tc.tilex = canvas.tilex + getCanvasShiftX();
-	tc.tiley = canvas.tiley + getCanvasShiftY();
-	tc.zoom = canvas.zoom;
-	tc.x = x2 + SCREEN_WIDTH / 2 + TILE_SIZE - canvas.x;
-	tc.y = y2 + SCREEN_HEIGHT / 2 + TILE_SIZE - canvas.y;
-
-	adjustTileCoordinateShifts(&tc);
-
-	WorldCoordinate wc;
-
-	toWorldCoordinate(&tc, &wc);
-
-	fprintf(stderr, "mousex = %d, mousey = %d, tilex=%d, tiley=%d, x=%.2f, y=%.2f, wc.lat=%.4f, wc.lon=%.4f\n", mouse.oldx, mouse.oldy, tc.tilex, tc.tiley, tc.x, tc.y, wc.latitude, wc.longitude);
-
-	GeoCodeResult* r = calloc(1, sizeof(GeoCodeResult));
-	r->centroid = wc;
-	r->searchId = canvas.lastSearchId;
-	addGeoCodeResult(r);
+	return closestMarker;
 }
 
 extern Uint8* keys;
 void handleMouseClick() {
-	checkClickedMarkers();
-	//	deallocateSearchMarkers();
+	ScreenMarker* marker = findMarkerUnderMouse();
 
-	if (keys[SDLK_LSHIFT]) {
-		addMarkerOnClickedPosition();
+	if (marker != NULL) {
+		setAttractionToPoint(&marker -> geoCodeResult -> centroid);
+		if (marker -> geoCodeResult -> description != NULL) {
+			addConsoleLine((char *) marker -> geoCodeResult -> description, 0, 1, 1);
+		}
+	}
+}
+
+void handleMouseUp() {
+	if (canvas.draggedMarker != NULL) {
+		canvas.draggedMarker -> raised = 0;
+		canvas.draggedMarker = NULL;
+	}
+	if (canvas.popup.popupEntries != NULL) {
+		if (canvas.popup.clickedPopupEntry != NULL) {
+			if (!canvas.popup.currentClickCreatedPopup) {
+				canvas.popup.clickedPopupEntry -> handlePressed(canvas.popup.clickedPopupEntry -> argument);
+				freePopupMenu();
+			} else if (canvas.popup.animationFrame >= 30) {
+				canvas.popup.clickedPopupEntry -> handlePressed(canvas.popup.clickedPopupEntry -> argument);
+				freePopupMenu();
+			}
+		}
+
+		if (canvas.popup.currentClickCreatedPopup) {
+			canvas.popup.currentClickCreatedPopup = FALSE;
+		} else {
+			freePopupMenu();
+		}
+	}
+	canvas.mouseHoldInPlace = -1;
+}
+
+void handleDraggedMarker() {
+	//		canvas.draggedMarker -> screenCoordinate.x += mouse.xdelta;
+	//		canvas.draggedMarker -> screenCoordinate.y += mouse.ydelta;
+	GLfloat markerMoveX = mouse.xdelta / canvas.scale;
+	GLfloat markerMoveY = mouse.ydelta / canvas.scale;
+
+	GLfloat canvasScrollX = 0.0;
+	GLfloat canvasScrollY = 0.0;
+
+	if (mouse.x < 100) {
+		canvasScrollX = .2;
+	}
+
+	if (mouse.x > SCREEN_WIDTH - 100) {
+		canvasScrollX = -.2;
+	}
+
+	if (mouse.y < 100) {
+		canvasScrollY = .2;
+	}
+
+	if (mouse.y > SCREEN_HEIGHT - 100) {
+		canvasScrollY = -.2;
+	}
+
+	if (canvasScrollX != 0.0 || canvasScrollY != 0.0) {
+		canvas.friction = 0.97;
+		GLfloat rotatedScrollX = canvasScrollY * sin(canvas.rotz * M_PI / 180.0) + canvasScrollX * sin((canvas.rotz + 90) * M_PI / 180.0);
+		GLfloat rotatedScrollY = canvasScrollY * cos(canvas.rotz * M_PI / 180.0) + canvasScrollX * cos((canvas.rotz + 90) * M_PI / 180.0);
+		canvas.dx += rotatedScrollX;
+		canvas.dy += rotatedScrollY;
+	}
+
+	rotate2d(&markerMoveX, &markerMoveY, -canvas.rotz);
+
+	canvas.draggedMarker -> tileCoordinate.x += markerMoveX - canvas.dx;
+	canvas.draggedMarker -> tileCoordinate.y += markerMoveY - canvas.dy;
+	adjustTileCoordinateShifts(&canvas.draggedMarker -> tileCoordinate);
+	toWorldCoordinate(&canvas.draggedMarker -> tileCoordinate, &canvas.draggedMarker -> geoCodeResult -> centroid);
+
+	if (canvas.draggedMarker -> type == ROUTE_START || canvas.draggedMarker -> type == ROUTE_END) {
+		if (canvas.route.query == NULL) {
+			canvas.route.query = calloc(1, sizeof(RoutingQuery));
+		}
+		if (canvas.draggedMarker -> type == ROUTE_START) {
+			canvas.route.query -> start = canvas.draggedMarker -> geoCodeResult -> centroid;
+		} else {
+			canvas.route.query -> end = canvas.draggedMarker -> geoCodeResult -> centroid;
+		}
+
+		canvas.routeMarkersChanged = TRUE;
+	}
+}
+
+PopupMenuEntry* createPopupEntry(void* argument, void(*handlePressed)(void*), char* displayName) {
+	PopupMenuEntry* entry = calloc(1, sizeof(PopupMenuEntry));
+	entry -> argument = argument;
+	entry -> displayName = displayName;
+	entry -> handlePressed = handlePressed;
+	return entry;
+}
+
+void directionsFromHere(void* worldCoordinatePtr) {
+	WorldCoordinate* wc = (WorldCoordinate*) worldCoordinatePtr;
+	if (canvas.route.query == NULL) {
+		canvas.route.query = calloc(1, sizeof(RoutingQuery));
+		canvas.route.query -> end = canvas.currentPos;
+	}
+	canvas.route.query -> start = *wc;
+	canvas.routeMarkersChanged = TRUE;
+
+	if (canvas.routeMarkers == NULL) {
+		createRouteMarkers();
+	} else {
+		updateRouteMarkers(ROUTE_START);
+	}
+
+}
+
+void directionsToHere(void* worldCoordinatePtr) {
+	WorldCoordinate* wc = (WorldCoordinate*) worldCoordinatePtr;
+	if (canvas.route.query == NULL) {
+		canvas.route.query = calloc(1, sizeof(RoutingQuery));
+		canvas.route.query -> start = canvas.currentPos;
+	}
+	canvas.route.query -> end = *wc;
+	canvas.routeMarkersChanged = TRUE;
+
+	if (canvas.routeMarkers == NULL) {
+		createRouteMarkers();
+	} else {
+		updateRouteMarkers(ROUTE_END);
+	}
+
+}
+
+void freePopupMenu() {
+	if (canvas.popup.popupEntries != NULL) {
+		g_list_free(canvas.popup.popupEntries);
+		canvas.popup.popupEntries = NULL;
+	}
+}
+
+void createSearchMarkerOnWorldCoordinate(void* wcPtr) {
+	WorldCoordinate* wc = wcPtr;
+	GeoCodeResult* gc = createGeoCodeResult(wc);
+	ScreenMarker* marker = createMarker(gc, SEARCH_RESULT);
+	canvas.markers = g_list_prepend(canvas.markers, marker);
+}
+
+void clearMap(void* unused) {
+	clearBuildings();
+	tileEngineFreeRoute();
+	deallocateSearchMarkers();
+	clearRouteMarkers();
+}
+
+void createPopupMenu() {
+	freePopupMenu();
+	TileCoordinate tc;
+	setClickedTileCoordinate(&tc);
+	WorldCoordinate* wc = calloc(1, sizeof(WorldCoordinate));
+	toWorldCoordinate(&tc, wc);
+
+	PopupMenuEntry* entry;
+
+	entry = createPopupEntry(wc, &directionsFromHere, "Directions from here");
+	canvas.popup.popupEntries = g_list_append(canvas.popup.popupEntries, entry);
+
+	entry = createPopupEntry(wc, &directionsToHere, "Directions to here");
+	canvas.popup.popupEntries = g_list_append(canvas.popup.popupEntries, entry);
+
+	entry = createPopupEntry(wc, &createSearchMarkerOnWorldCoordinate, "Create marker here");
+	canvas.popup.popupEntries = g_list_append(canvas.popup.popupEntries, entry);
+
+	entry = createPopupEntry(wc, &clearMap, "Clear map");
+	canvas.popup.popupEntries = g_list_append(canvas.popup.popupEntries, entry);
+
+	canvas.popup.changed = TRUE;
+	canvas.popup.currentClickCreatedPopup = TRUE;
+}
+
+void handleMouseHoldOnMap() {
+	TileCoordinate tc;
+	if (fabs(canvas.dx) > 0.1 || fabs(canvas.dy) > 0.1) {
+		return;
+	}
+
+	if (canvas.mouseHoldInPlace < 0) {
+		setClickedTileCoordinate(&tc);
+		toWorldCoordinate(&tc, &canvas.mouseHoldPositionWc);
+	}
+
+	toTileCoordinate(&canvas.mouseHoldPositionWc, &tc, canvas.zoom);
+	toScreenCoordinate(&tc, &canvas.mouseHoldPositionSc);
+
+	canvas.mouseHoldInPlace = nowMillis - mouse.pressed;
+
+	if (canvas.popup.popupEntries == NULL && canvas.mouseHoldInPlace > options.mouseHoldToPopupTime) {
+		createPopupMenu();
+	}
+}
+
+void handleMouseHold() {
+	ScreenMarker* marker = findMarkerUnderMouse();
+	int mouseHoldInPlace = FALSE;
+
+	if (mouse.draggedDistanceX < 15 && mouse.draggedDistanceY < 15) {
+		mouseHoldInPlace = TRUE;
+	}
+
+	if (mouseHoldInPlace) {
+		if (canvas.draggedMarker == NULL && marker != NULL) {
+			marker -> raised = 1;
+			canvas.draggedMarker = marker;
+		}
+	}
+
+	if (canvas.draggedMarker != NULL) {
+		handleDraggedMarker();
+	} else if (mouseHoldInPlace) {
+		handleMouseHoldOnMap();
+	} else {
+		canvas.mouseHoldInPlace = -1;
+	}
+}
+
+GLfloat getPopupEntryHeight();
+void updateClickedPopupEntry() {
+	if (canvas.popup.popupEntries == NULL) {
+		canvas.popup.clickedPopupEntry = NULL;
+		return;
+	}
+	int x, y;
+
+	TileCoordinate tc;
+	Point clickedPoint;
+	setClickedTileCoordinate(&tc);
+	toScreenCoordinate(&tc, &clickedPoint);
+
+	x = clickedPoint.x - canvas.popup.x;
+	y = clickedPoint.y - canvas.popup.y;
+	if (x < 0 || y < 0 || x > canvas.popup.width || y >= canvas.popup.height) {
+		canvas.popup.clickedPopupEntry = NULL;
+		return;
+	}
+	int idx = y / getPopupEntryHeight();
+	int length = g_list_length(canvas.popup.popupEntries);
+	if (idx < 0 || idx >= length) {
+		fprintf(stderr, "Internal error: clicked popup index out of range, popupSize = %d, index = %d\n", length, idx);
+		canvas.popup.clickedPopupEntry = NULL;
+		return;
+	}
+	canvas.popup.clickedPopupEntry = g_list_nth(canvas.popup.popupEntries, idx) -> data;
+}
+
+void tileEngineHandleMouseDown() {
+	if (canvas.popup.popupEntries != NULL) {
+		if (canvas.popup.animationFrame >= 30) {
+			if (canvas.popup.clickedPopupEntry == NULL) {
+				freePopupMenu();
+			}
+		}
 	}
 }
 
@@ -856,19 +1238,30 @@ void tileEngineProcessMouse(int uiElemPressed) {
 		return;
 	}
 
+	updateClickedPopupEntry();
 	if (mouse.button != 0) {
+		if (mouse.oldButton == 0) {
+			tileEngineHandleMouseDown();
+		}
 		canvas.followingMypos = 0;
+		canvas.routeFlybyMode = 0;
 		canvas.attractionToPoint = 0;
 		if (canvas.viewMode == VIEW_3D || canvas.rotation2d == 1) {
 			canvas.attractedToRotZ = 0;
 		}
+		if (mouse.oldButton != 0 && nowMillis - mouse.pressed > 200) {
+			handleMouseHold();
+		}
 	} else {
 		if (mouse.oldButton != 0 && nowMillis - mouse.pressed < 200) {
 			handleMouseClick();
 		}
+		if (mouse.oldButton != 0) {
+			handleMouseUp();
+		}
 	}
 
-	if (canvas.viewMode == VIEW_3D || canvas.rotation2d == 1) {
+	if ((canvas.viewMode == VIEW_3D || canvas.rotation2d == 1) && canvas.draggedMarker == NULL) {
 		if (options.orientation == PORTRAIT && mouse.x < 150) {
 			canvas.drotz -= mouse.ydelta / 18.0;
 			return;
@@ -879,12 +1272,11 @@ void tileEngineProcessMouse(int uiElemPressed) {
 		}
 	}
 
-	GLfloat dx = mouse.ydelta * sin(canvas.rotz * M_PI / 180.0) + mouse.xdelta * sin((canvas.rotz + 90) * M_PI / 180.0);
-	GLfloat dy = mouse.ydelta * cos(canvas.rotz * M_PI / 180.0) + mouse.xdelta * cos((canvas.rotz + 90) * M_PI / 180.0);
-
-	if (mouse.button) {
-		canvas.dx += dx / 10.0;
-		canvas.dy += dy / 10.0;
+	if (mouse.button && canvas.draggedMarker == NULL && canvas.popup.popupEntries == NULL) {
+		GLfloat dx = mouse.ydelta * sin(canvas.rotz * M_PI / 180.0) + mouse.xdelta * sin((canvas.rotz + 90) * M_PI / 180.0);
+		GLfloat dy = mouse.ydelta * cos(canvas.rotz * M_PI / 180.0) + mouse.xdelta * cos((canvas.rotz + 90) * M_PI / 180.0);
+		canvas.dx += dx / 10.0 / canvas.scale;
+		canvas.dy += dy / 10.0 / canvas.scale;
 		if (nowMillis - mouse.moved > 50) {
 			canvas.friction = .2;
 		} else {
@@ -1011,6 +1403,7 @@ void tileEngineToggleSearchBar() {
 	//		}
 }
 
+double prevDistance = 0;
 void tileEngineUpdateCoordinates() {
 	//int zoomChange = (int) round(log2((double) canvas.scale));
 	int zoomChange = 0;
@@ -1042,6 +1435,86 @@ void tileEngineUpdateCoordinates() {
 		canvas.attraction = canvas.currentPos;
 		canvas.attractionToPoint = 1;
 #endif
+	} else if (canvas.routeFlybyMode == 1) {
+		if (canvas.route.optimizedRoute.lineStripTileCoordinates == NULL) {
+			canvas.routeFlybyMode = 0;
+		} else {
+			TileCoordinate tc;
+			WorldCoordinate wc;
+			tc = g_array_index(canvas.route.optimizedRoute.lineStripTileCoordinates, TileCoordinate, 0);
+			toWorldCoordinate(&tc, &wc);
+			canvas.attraction = wc;
+			canvas.attractionToPoint = 1;
+
+			if (getDistance(&canvas.attraction, &canvas.center) < 1.0) {
+				canvas.routeFlybyMode = 2;
+				canvas.routeFlybyLineIdx = 0;
+				canvas.dx = 0;
+				canvas.dy = 0;
+			}
+		}
+	} else if (canvas.routeFlybyMode == 2) {
+		if (canvas.route.optimizedRoute.lineStripTileCoordinates == NULL) {
+			canvas.routeFlybyMode = 0;
+		} else {
+			TileCoordinate nextRoutePointTc;
+
+			double distance = getDistance(&canvas.center, &canvas.center);
+
+			if ((prevDistance >= 0 && distance >= prevDistance) || canvas.routeFlybyLineIdx == 0) {
+				if (canvas.routeFlybyLineIdx < canvas.route.optimizedRoute.lineStripTileCoordinates -> len - 1) {
+					if (canvas.routeFlybyLineIdx <= canvas.navigationStatus.currentLineIdx + 1) {
+						TileCoordinate prevRoutePointTc = g_array_index(canvas.route.optimizedRoute.lineStripTileCoordinates,
+													TileCoordinate, canvas.routeFlybyLineIdx);
+						canvas.routeFlybyLineIdx++;
+						nextRoutePointTc = g_array_index(canvas.route.optimizedRoute.lineStripTileCoordinates,
+								TileCoordinate, canvas.routeFlybyLineIdx);
+
+						if (canvas.viewMode == VIEW_3D || canvas.rotation2d) {
+
+							Point prevSc, nextSc;
+							toScreenCoordinate(&prevRoutePointTc, &prevSc);
+							toScreenCoordinate(&nextRoutePointTc, &nextSc);
+							Point e; // east
+							e.x = nextSc.x - prevSc.x;
+							e.y = nextSc.y - prevSc.y;
+
+							if (e.x != 0.0) {
+								Point zero;
+								zero.x = 0.0;
+								zero.y = 0.0;
+								double length = pointDistance(&zero, &e);
+								e.x /= length;
+								e.y /= length;
+								double theta = atan2(e.y, e.x);
+								canvas.destRotz = -theta / M_PI * 180.0;
+								if (options.orientation == LANDSCAPE) {
+									canvas.destRotz -= 90.0;
+									if (canvas.destRotz < -180) {
+										canvas.destRotz += 360;
+									}
+								} else {
+									canvas.destRotz += 180.0;
+									if (canvas.destRotz > 180) {
+										canvas.destRotz -= 360;
+									}
+								}
+								canvas.attractedToRotZ = 1;
+							}
+						}
+						toWorldCoordinate(&nextRoutePointTc, &canvas.attraction);
+						prevDistance = -1;
+					}
+				} else {
+					canvas.routeFlybyMode = 0;
+//					canvas.attractionToPoint = 0;
+//					canvas.dx = 0;
+//					canvas.dy = 0;
+				}
+			} else {
+				prevDistance = distance;
+			}
+		}
 	}
 
 	if (canvas.attractionToPoint) {
@@ -1093,8 +1566,17 @@ void tileEngineUpdateCoordinates() {
 
 		canvas.friction = .99;
 
-		canvas.dx = (tmp.x / 20.0) / canvas.scale;
-		canvas.dy = (tmp.y / 20.0) / canvas.scale;
+		if (canvas.routeFlybyMode == 2) {
+			Point p;
+			p.x = 0.0;
+			p.y = 0.0;
+			double dist = pointDistance(&p, &tmp);
+			canvas.dx = tmp.x * 3.0 / dist / canvas.scale;
+			canvas.dy = tmp.y * 3.0 / dist / canvas.scale;
+		} else {
+			canvas.dx = (tmp.x / 20.0) / canvas.scale;
+			canvas.dy = (tmp.y / 20.0) / canvas.scale;
+		}
 
 		if (canvas.destScale == 0.0 && canvas.attractionZooming == 1) {
 			canvas.attractionZooming = 0;
@@ -1305,6 +1787,7 @@ void tileEngineGotomypos() {
 	if (mouse.oldButton == 0) {
 		canvas.followingMypos = 1 - canvas.followingMypos;
 		canvas.attractionZooming = 0;
+		canvas.routeFlybyMode = 0;
 		if (canvas.followingMypos) {
 			updateCanvasCenterWorldCoordinate();
 			canvas.previousCenter.latitude = canvas.center.latitude;
@@ -1357,12 +1840,20 @@ void setZoomBarInitialized(int value) {
 	zoomBarInitialized = value;
 }
 
-void tmpAddMarker() {
-	GeoCodeResult* r = calloc(1, sizeof(GeoCodeResult));
-	if (r != NULL) {
-		r -> centroid = canvas.center;
-		r-> searchId = canvas.lastSearchId;
-		addGeoCodeResult(r);
+void addSearchMarkerAtCanvasCenter() {
+	GeoCodeResult* r = createGeoCodeResult(&canvas.center);
+	ScreenMarker* marker = createMarker(r, SEARCH_RESULT);
+	canvas.markers = g_list_prepend(canvas.markers, marker);
+}
+
+void tmpQueryRoute() {
+	RoutingQuery* q = calloc(1, sizeof(RoutingQuery));
+	if (q != NULL) {
+		q -> start.latitude = canvas.currentPos.latitude;
+		q -> start.longitude = canvas.currentPos.longitude;
+		q -> end.latitude = canvas.center.latitude;
+		q -> end.longitude = canvas.center.longitude;
+		processNewRoutingQuery(q);
 	}
 }
 
@@ -1530,12 +2021,12 @@ void tileEngine_computeTilesVisibility() {
 					if (tiles[i][j][currentTilesIdx] -> state == STATE_GL_TEXTURE_NOT_CREATED && texturesCreatedThisFrame < MAX_TEXTURES_CREATED_PER_FRAME) {
 						tiles[i][j][currentTilesIdx]->texture = createTexture(tiles[i][j][currentTilesIdx] -> pixels4444, TILE_SIZE, TILE_SIZE, TRUE);
 						// TODO tmp
-//						if (syncLoadedTilesAppearance == TRUE) {
-//							tiles[i][j][currentTilesIdx] -> state = STATE_LOADED_SYNC_WAIT;
-//						} else {
-							tiles[i][j][currentTilesIdx] -> state = STATE_LOADED;
-							tiles[i][j][currentTilesIdx] ->stateChangeTime = 0;
-//						}
+						//						if (syncLoadedTilesAppearance == TRUE) {
+						//							tiles[i][j][currentTilesIdx] -> state = STATE_LOADED_SYNC_WAIT;
+						//						} else {
+						tiles[i][j][currentTilesIdx] -> state = STATE_LOADED;
+						tiles[i][j][currentTilesIdx] ->stateChangeTime = 0;
+						//						}
 						texturesCreatedThisFrame++;
 					}
 					tiles[i][j][currentTilesIdx] -> visible = TRUE;
